/*
 * Copyright 2011 Cedric Duval <cedricduval@free.fr>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2.
 *
 * See the accompanying COPYING file for more details.
 */

/**
 * \file
 * Read a test file and generate the corresponding shell commands
 *
 * These commands, when executed by a bourne shell, will generate an
 * output suitable for reading with an istream.
 */

#include "config.h"

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <stdbool.h>
#include <string.h>
#include <assert.h>

#include "cmdgen.h"
#include "wbuf.h"


enum state {
	STATE_COMMAND,  /* within a command or command continuation */
	STATE_OUTPUT,   /* within command output */
	STATE_NONE,     /* neither command nor output */
	STATE_END       /* last line of input has been processed */
};

struct linebuf {
	char *data;
	size_t len;
};

struct cmdgen_context {
	enum state state;       /* current parsing state */
	enum state prev_state;  /* previous parsing state */

	unsigned int indent;    /* indentation size */

	unsigned int keep;      /* number of command and comment lines */
	unsigned int skip;      /* number of command output lines */

	char *salt;             /* random salt string */

	struct linebuf *buf;    /* line buffer */
	struct wbuf *cmd;       /* command lines buffer */

	FILE *in;               /* input stream */
	FILE *out;              /* output stream */
};



/* --------------------------------------------------------------------------
 *   private functions
 */

static void
emit_salt(struct cmdgen_context *c)
{
	static const char * const states = "*=#.";

	c->keep -= c->skip;

	fprintf(c->out, "echo \"%s $? %d %d %c%c @\"\n", c->salt,
	                                                 c->keep,
	                                                 c->skip,
	                                                 states[c->prev_state],
	                                                 states[c->state]);

	c->keep = 0;
	c->skip = 0;
}

static void
write_header(struct cmdgen_context *c)
{
	fprintf(c->out, "echo \"%s %s %u\"\n", "|klp|", c->salt, c->indent);
}

static inline void
switch_state(struct cmdgen_context *c, int state)
{
	c->prev_state = c->state;
	c->state = state;
}

static int
process_line(struct cmdgen_context *c, const char *line)
{
	const char *p, *start;

	if (!c->indent) {
		/* detect indentation
		 */
		start = strchr(line, '$');
		if (!start || start == line)
			return 0;
		for (p = line; p != start; p++)
			if (*p != ' ')
				return 0;
		c->indent = start - line;
		write_header(c);
	}

	start = line + c->indent;
	for (p = line; *p && p != start; p++)
		if (*p != ' ')
			break;

	if (p != start) {
		/* line with no indentation
		 */
		switch (c->state) {
		case STATE_COMMAND:
		case STATE_OUTPUT:
			switch_state(c, STATE_NONE);
			emit_salt(c);
			return wbuf_print(c->cmd);
		case STATE_NONE:
			return 0;
		default:
			assert(0);
			return -1;
		}
	} else if (*start == '$' && start[1] == ' ') {
		/* command prefix
		 */
		switch (c->state) {
		case STATE_COMMAND:
			emit_salt(c);
			if (wbuf_print(c->cmd) == -1)
				return -1;
			return wbuf_append(c->cmd, start + 2);
		case STATE_OUTPUT:
			emit_salt(c);
			if (wbuf_print(c->cmd) == -1)
				return -1;
			/* fall through */
		case STATE_NONE:
			switch_state(c, STATE_COMMAND);
			return wbuf_append(c->cmd, start + 2);
		default:
			assert(0);
			return -1;
		}
	} else if (*start++ == '>' && (*start == ' ' || *start == '\n')) {
		/* command continuation prefix
		 */
		switch (c->state) {
		case STATE_COMMAND:
			if (*start != '\n')
				++start;
			return wbuf_append(c->cmd, start);
		case STATE_OUTPUT:
			switch_state(c, STATE_NONE);
			emit_salt(c);
			return wbuf_print(c->cmd);
		case STATE_NONE:
			return 0;
		default:
			assert(0);
			return -1;
		}
	} else {
		/* output prefix
		 */
		switch (c->state) {
		case STATE_COMMAND:
			switch_state(c, STATE_OUTPUT);
			/* fall through */
		case STATE_OUTPUT:
			c->skip++;
			/* fall through */
		case STATE_NONE:
			return 0;
		default:
			assert(0);
			return -1;
		}
	}

	assert(0);
	return -1;
}

static char *
generate_salt(char *buf, size_t len)
{
	FILE *fp;
	char *p;

	p = getenv("KUULTO_SALT");
	if (p && *p) /* environment override */
		return p;

	fp = fopen("/dev/urandom", "r");
	if (!fp) {
		snprintf(buf, len, "SALT%d%d", getpid(), getppid());
		return buf;
	}
	p = buf;
	while (--len)
		*p++ = ((fgetc(fp) % 2) ? 'a' : 'A') + (fgetc(fp) % 26);
	*p = 0;
	fclose(fp);

	return buf;
}



/* --------------------------------------------------------------------------
 *   public functions
 */

struct cmdgen_context *
cmdgen_init(FILE *in, FILE *out)
{
	struct cmdgen_context *c;
	struct linebuf *lbuf;
	struct wbuf *cmd;
	char saltbuf[12];
	char *salt;

	cmd = wbuf_init(out);
	if (!cmd)
		return NULL;

	if ( !(c    = calloc(1, sizeof(*c)))   ||
	     !(lbuf = malloc(sizeof(*c->buf))) ||
	     !(salt = strdup(generate_salt(saltbuf, sizeof(saltbuf)))) )
	{
		fprintf(stderr, "Out of memory!\n");
		if (c) {
			if (lbuf)
				free(salt);
			free(lbuf);
			free(c);
		}
		return NULL;
	}

	c->state      = STATE_NONE;
	c->indent     = 0;
	c->keep       = 0;
	c->skip       = 0;
	c->cmd        = cmd;
	c->salt       = salt;
	c->buf        = lbuf;
	c->buf->data  = 0;
	c->buf->len   = 0;
	c->in         = in;
	c->out        = out;

	return c;
}

/* returns the number of line processed (0 indicates end of input)
 * or -1 in case of error
 * 
 * memory and input stream both implicitly released when processing is over
 */
int
cmdgen_process_line(struct cmdgen_context *c)
{
	struct linebuf *buf = c->buf;
	int res = -1;

	if (getline(&buf->data, &buf->len, c->in) != -1) {
		if (process_line(c, buf->data) == -1)
			goto out;

		c->keep++;
		return 1; /* one line processed */
	}

	/* finalize */

	if (!c->indent)
		/* no command found, but there might be content still */
		write_header(c);

	c->prev_state = c->state;
	c->state = STATE_END;
	emit_salt(c);
	res = (wbuf_print(c->cmd) == -1) ? -1 : 0;

	c->prev_state = c->state;
	emit_salt(c);

out:
	fclose(c->in);
	fclose(c->out); /* eof to consumer */
	wbuf_free(c->cmd);
	free(c->buf->data);
	free(c->buf);
	free(c->salt);
	free(c);

	return res;
}

int
cmdgen_process_stream(FILE *in, FILE *out)
{
	struct cmdgen_context *c;
	int res;

	c = cmdgen_init(in, out);
	if (!c)
		return -1;

	while ( (res = cmdgen_process_line(c)) > 0)
		;

	return res;
}
