/*
 * Copyright 2011 Cedric Duval <cedricduval@free.fr>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2.
 *
 * See the accompanying COPYING file for more details.
 */

/**
 * \file
 * Various file descriptors utilities
 *
 * Essentially, this provides wrappers for setting the close-and-exec flag
 * in a race-free manner whenever the kernel and C library support it
 * (for instance using POSIX.1-2008.'s O_CLOEXEC for open(2)), and falling
 * back to fcntl(2) otherwise.
 *
 * We don't just assume that because a flag is defined, the functionality
 * is operational: the underlying kernel might not know that flag, so on
 * the first call we check whether it actually works, and then decide how
 * it will be handled from there on.
 *
 * It also provides crude replacements for POSIX.1-2008 -at functions
 * (openat, etc.) when the system lacks them.
 */

#include "config.h"

#ifdef HAVE_PIPE2
# define _GNU_SOURCE
#endif
#include <pthread.h>
#include <stdio.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <stdbool.h>
#include <string.h>
#include <errno.h>
#include <assert.h>


#include "fdutils.h"

/* in the fallback-to-fcntl case, the whole sequence (file descriptor
 * creation plus setting of the close-on-exec flag) is locked by a mutex
 * also used by fork(2), to ensure another child process won't be created
 * in the meantime */
pthread_mutex_t fdutils_mutex = PTHREAD_MUTEX_INITIALIZER;


/* --------------------------------------------------------------------------
 *   private functions
 */

/* using fcntl(2) as last resort if other better means are not supported */
static int
set_cloexec(int fd)
{
	int flags;

	flags = fcntl(fd, F_GETFD);
	if (flags == -1) {
		perror("fcntl get");
		return -1;
	}

	if (fcntl(fd, F_SETFD, flags | FD_CLOEXEC) == -1) {
		perror("fcntl set");
		return -1;
	}

	return 0;
}

/* called after a first attempt to set close-on-exec directly, reports
 * whether it is actually supported by the kernel/libc combination */
static int
check_cloexec(int fd)
{
	int flags;

	flags = fcntl(fd, F_GETFD);
	if (flags == -1) {
		perror("fcntl get");
		return -1;
	}

	return (flags & FD_CLOEXEC) ? 1 : -1;
}

#if !defined(HAVE_FSTATAT) || !defined(HAVE_MKDIRAT) || !defined(HAVE_OPENAT)
/* This potentially doesn't cover compatibility with all systems, but this
 * will do until an actual user shows up, of a platform which is neither
 * standards-compliant nor supports /proc/self/fd */
static const char * const support_msg =
  "POSIX-1.2008 -at functions not properly supported\n"
  "Please report to " PACKAGE_BUGREPORT "\n";

static int
path_at(int dir_fd, const char *path, char *buf, size_t len)
{
	char procbuf[64];
	ssize_t n;

	/* determine the path to the directory associated with dir_fd */
	snprintf(procbuf, sizeof(procbuf), "/proc/self/fd/%d", dir_fd);
	n = readlink(procbuf, buf, len);
	if (n == -1) {
		fprintf(stderr, "readlink: %s: %s\n", procbuf, strerror(errno));
		fprintf(stderr, "%s", support_msg);
		return -1;
	}

	/* concatenate the relative path */
	if (snprintf(buf+n, sizeof(buf)-n, "/%s", path) >= n) {
		errno = ENOBUFS;
		return -1;
	}

	return 0;
}
#endif /* !defined(HAVE_FSTATAT) || !defined(HAVE_MKDIRAT) */


/* --------------------------------------------------------------------------
 *   public functions
 */

int
open_cloexec(const char *pathname, int flags)
{
	int fd;
	bool ok;

#ifdef O_CLOEXEC
	static int cloexec_ok; /* -1 unsupported, 0 unknown, 1 supported */

restart:
	if (cloexec_ok >= 0) {
		/* using POSIX.1-2008. O_CLOEXEC */
		fd = open(pathname, flags | O_CLOEXEC);
		if (fd == -1) {
			fprintf(stderr, "open: %s: %s\n", pathname,
			                                  strerror(errno));
			return -1;
		}

		if (cloexec_ok == 0)
			cloexec_ok = check_cloexec(fd);

		if (cloexec_ok < 0) {
			close(fd);
			goto restart; /* restart, with locked fallback */
		}

		return fd;
	}
#endif

	/* fall back to open(2)+fcntl(2)... */
	pthread_mutex_lock(&fdutils_mutex);
	fd = open(pathname, flags);
	if (fd == -1) {
		fprintf(stderr, "open: %s: %s\n", pathname, strerror(errno));
		pthread_mutex_unlock(&fdutils_mutex);
		return -1;
	}
	ok = set_cloexec(fd) != -1;
	pthread_mutex_unlock(&fdutils_mutex);

	return ok ? fd : -1;
}

int
openat_cloexec(int dirfd, const char *pathname, int flags)
{
	int fd;
	bool ok;

#ifdef O_CLOEXEC
	static int cloexec_ok; /* -1 unsupported, 0 unknown, 1 supported */

restart:
	if (cloexec_ok >= 0) {
		/* using POSIX.1-2008. O_CLOEXEC */
		fd = xopenat(dirfd, pathname, flags | O_CLOEXEC);
		if (fd == -1) {
			fprintf(stderr, "xopenat: %s: %s\n", pathname,
			                                     strerror(errno));
			return -1;
		}

		if (cloexec_ok == 0)
			cloexec_ok = check_cloexec(fd);

		if (cloexec_ok < 0) {
			close(fd);
			goto restart; /* restart, with locked fallback */
		}

		return fd;
	}
#endif

	/* fall back to openat(2)+fcntl(2)... */
	pthread_mutex_lock(&fdutils_mutex);
	fd = xopenat(dirfd, pathname, flags);
	if (fd == -1) {
		fprintf(stderr, "xopenat: %s: %s\n", pathname, strerror(errno));
		pthread_mutex_unlock(&fdutils_mutex);
		return -1;
	}
	ok = set_cloexec(fd) != -1;
	pthread_mutex_unlock(&fdutils_mutex);

	return ok ? fd : -1;
}

int
pipe_cloexec(int pipefd[2])
{
	bool ok;

#ifdef HAVE_PIPE2
	static int cloexec_ok; /* -1 unsupported, 0 unknown, 1 supported */

	if (cloexec_ok >= 0) {
		if (pipe2(pipefd, O_CLOEXEC) == -1) {
			perror("pipe2");
			return -1;
		}

		if (cloexec_ok == 0)
			cloexec_ok = check_cloexec(pipefd[0]);

		if (cloexec_ok > 0)
			return 0; /* close-on-exec successfully set */
	}
	else /* fall back to pipe(2)+fcntl(2)... */
#endif
	{
		pthread_mutex_lock(&fdutils_mutex);
		if (pipe(pipefd) == -1) {
			perror("pipe");
			pthread_mutex_unlock(&fdutils_mutex);
			return -1;
		}
	}

	ok = (set_cloexec(pipefd[0]) != -1 && set_cloexec(pipefd[1]) != -1);
	pthread_mutex_unlock(&fdutils_mutex);

	return ok ? 0 : -1;
}

FILE *
fopen_cloexec(const char *path, const char *mode)
{
	static int cloexec_ok; /* -1 unsupported, 0 unknown, 1 supported */
	FILE *fp;
	bool ok;

restart:
	if (cloexec_ok >= 0) {
		/* using the 'e' flag to fopen to specify close-on-exec */
		size_t modelen = strlen(mode);
		char bufmode[8], *p;

		assert(modelen < sizeof(bufmode) - 1);

		memcpy(bufmode, mode, modelen);
		p = bufmode + modelen;
		*p++ = 'e';
		*p = 0;

		fp = fopen(path, bufmode);
		if (!fp) {
			if (errno == EINVAL) {
				cloexec_ok = -1;
				goto restart;
			} else {
				fprintf(stderr, "fopen: %s: %s\n",
				                path, strerror(errno));
				return NULL;
			}
		}

		if (cloexec_ok == 0)
			cloexec_ok = check_cloexec(fileno(fp));

		if (cloexec_ok < 0) {
			fclose(fp);
			goto restart; /* restart, with locked fallback */
		}

		return fp; /* close-on-exec successfully set */
	}

	/* fall back to fopen(2)+cntl(2)... */
	pthread_mutex_lock(&fdutils_mutex);
	fp = fopen(path, mode);
	if (!fp) {
		fprintf(stderr, "fopen: %s: %s\n", path, strerror(errno));
		pthread_mutex_unlock(&fdutils_mutex);
		return NULL;
	}
	ok = set_cloexec(fileno(fp)) != -1;
	pthread_mutex_unlock(&fdutils_mutex);

	return ok ? fp : NULL;
}

#ifndef HAVE_FSTATAT
int
xfstatat(int dir_fd, const char *path, struct stat *st, int flag)
{
	char buf[4096];

	if (path_at(dir_fd, path, buf, sizeof(buf)) == -1)
		return -1;

	return stat(buf, st);
}
#endif /* !HAVE_FSTATAT */

#ifndef HAVE_MKDIRAT
int
xmkdirat(int dir_fd, const char *path, mode_t mode)
{
	char buf[4096];

	if (path_at(dir_fd, path, buf, sizeof(buf)) == -1)
		return -1;

	return mkdir(buf, mode);
}
#endif /* !HAVE_MKDIRAT */

#ifndef HAVE_OPENAT
int
xopenat(int dir_fd, const char *pathname, int flags)
{
	char buf[4096];

	if (path_at(dir_fd, pathname, buf, sizeof(buf)) == -1)
		return -1;

	return open(buf, flags);
}
#endif /* !HAVE_OPENAT */
