/*
 * Copyright 2011 Cedric Duval <cedricduval@free.fr>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2.
 *
 * See the accompanying COPYING file for more details.
 */

/**
 * \file
 * Various file descriptors utilities
 *
 * Essentially, this provides wrappers for setting the close-and-exec flag
 * in a race-free manner whenever the kernel and C library support it
 * (for instance using POSIX.1-2008.'s O_CLOEXEC for open(2)), and falling
 * back to fcntl(2) otherwise.
 *
 * It also provides crude replacements for POSIX.1-2008 -at functions
 * (openat, etc.) when the system lacks them.
 */

#ifndef _KUULTO_FDUTILS_H
#define _KUULTO_FDUTILS_H

/**
 * Open a file/directory with the close-on-exec flag set
 *
 * See open(2) for parameters
 */
int open_cloexec(const char *pathname, int flags);

/**
 * Open a file/directory with the close-on-exec flag set
 *
 * See openat(2) for parameters
 */
int openat_cloexec(int dirfd, const char *pathname, int flags);

/**
 * Open a file with the close-on-exec flag set
 *
 * See fopen(3) for parameters
 */
FILE *fopen_cloexec(const char *path, const char *mode);

/**
 * Create a pipe with the close-on-exec flag set
 *
 * See pipe(2) for parameters
 */
int pipe_cloexec(int pipefd[2]);

# include <sys/stat.h>
#ifdef HAVE_FSTATAT
# define xfstatat(dir_fd, path, st, flag)  fstatat(dir_fd, path, st, flag)
#else
int xfstatat(int dir_fd, const char *path, struct stat *st, int flag);
#endif /* HAVE_FSTATAT */

#ifdef HAVE_MKDIRAT
# include <sys/stat.h>
# define xmkdirat(dir_fd, path, mode)  mkdirat(dir_fd, path, mode)
#else
int xmkdirat(int dir_fd, const char *path, mode_t mode);
#endif /* HAVE_MKDIRAT */

#ifdef HAVE_OPENAT
# include <fcntl.h>
# define xopenat(dir_fd, pathname, flags)  openat(dir_fd, pathname, flags)
#else
int xopenat(int dir_fd, const char *pathname, int flags);
#endif /* HAVE_OPENAT */

#endif /* _KUULTO_FDUTILS_H */
