/*
 * Copyright 2011 Cedric Duval <cedricduval@free.fr>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2.
 *
 * See the accompanying COPYING file for more details.
 */

/**
 * \file
 * Support for merging a test result back into a test file
 */

#ifndef _KUULTO_MERGE_H
#define _KUULTO_MERGE_H

/* --------------------------------------------------------------------------
 *   flags
 */

/**
 * Disable escaping and regexp/glob support
 */
#define MERGE_FLAG_VERBATIM  (1 << 0)



/* --------------------------------------------------------------------------
 *   functions
 */

/**
 * Allocate a merging session
 *
 * \param[in] fp_test    Test input stream
 * \param[in] fp_result  Result stream: shell output
 * \param[in] fp_out     Merged output stream
 * \param[in] flags      Flags to be applied
 *
 * \return               Handle suitable for use with merge_init() and
 *                       merge_process_input(), or NULL in case of error
 */
struct merge_context *merge_alloc(FILE *fp_test,
                                  FILE *fp_result,
                                  FILE *fp_out,
                                  unsigned int flags);

/**
 * Initialize a merging session
 *
 * Initialization is a separate step to allow allocating a merging context
 * while the input streams are not yet ready to be read (without blocking).
 *
 * \param[in] c  Handle supplied by merge_init()
 *
 * \return       O if the result stream may be processed, -1 in case of error
 */
int merge_init(struct merge_context *c);

/**
 * Process one line of results and merge it with the original test
 *
 * The memory and input streams associated to the merging session are
 * implicitly released when processing is over.
 *
 * \param[in] c  Handle supplied by merge_init()
 *
 * \return       Number of line processed (0 indicates end of processing)
 *               or -1 in case of error
 */
int merge_process_result_line(struct merge_context *c);

/**
 * Process the entirety of the input streams
 *
 * The memory and input streams associated to the merging session are
 * implicitly released when processing is over.
 *
 * \param[in] fp_test    Test input stream
 * \param[in] fp_result  Result stream: shell output
 * \param[in] fp_out     Merged output stream
 * \param[in] flags      Flags to be applied
 *
 * \return               0 if processing succeeded, -1 otherwise
 */
int merge_process_streams(FILE *fp_test,
                          FILE *fp_result,
                          FILE *fp_out,
                          unsigned int flags);

#endif /* _KUULTO_MERGE_H */
