/*
 * Copyright 2011 Cedric Duval <cedricduval@free.fr>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2.
 *
 * See the accompanying COPYING file for more details.
 */

#include "config.h"

#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <stdbool.h>
#include <errno.h>

#include "../merge.h"
#include "../tmpdir.h"

#define RC_ERROR  255

int run_test(const char *path, FILE *out, unsigned int merge_flags, int *rc);


static const char * const usage_string =
  "\n"
  "Usage: %s [ <options> ] <test_file>\n"
  "\n"
  "This program runs a single test file.\n"
  "\n"
  "Options:\n"
  " -h         show this help text\n"
  " -k         don't remove the temporary directory used for tests\n"
  "            (alternatively, the path of a directory to use may be\n"
  "            specified through the KUULTO_TMPDIR environment variable)\n"
  " -o <path>  file receiving merged output (stdout if none specified)\n"
  " -q         quiet mode, no visual delimiters around output\n"
  " -V         verbatim mode, disable escaping and regexp/glob support\n"
  "";



/* --------------------------------------------------------------------------
 *   private functions
 */

static int
usage(const char *argv0, const char *err)
{
	int rc = (err && *err) ? RC_ERROR : 0;

	fprintf(stderr, usage_string, argv0);

	if (rc != 0)
		fprintf(stderr, "\nError: %s\n", err);

	return rc;
}



/* --------------------------------------------------------------------------
 *   public functions
 */

int
main(int argc, char **argv)
{
	static const char * const d = "________________________";
	const char *p, *in_file, *out_file = NULL;
	unsigned int merge_flags = 0;
	bool quiet = false;
	int opt, res, rc;
	FILE *out;

	while ((opt = getopt(argc, argv, "hko:qV")) != -1)
		switch (opt) {
		case 'h': return usage(argv[0], "");
		case 'k': tmpdir_keep(); break;
		case 'o': out_file = optarg; break;
		case 'q': quiet = true; break;
		case 'V': merge_flags |= MERGE_FLAG_VERBATIM; break;
		default: return usage(argv[0], "Invalid option");
		}

	if (argc - optind != 1)
		return usage(argv[0], "");
	in_file = argv[optind];

	if (!out_file) {
		out = stdout;
	} else {
		out = fopen(out_file, "w");
		if (!out) {
			perror("fopen");
			return RC_ERROR;
		}
	}

	if (!quiet) {
		/* header */
		p = strrchr(in_file, '/');
		p = p ? p + 1 : in_file;
		fprintf(out, " %s%s%s\n/%72s\n", d, d, d, p);
	}

	res = run_test(in_file, out, merge_flags, &rc);

	if (res != -1 && !quiet)
		/* footer */
		fprintf(out, "\\%s%s%s\n", d, d, d);

	return (res != -1) ? rc : RC_ERROR;
}
